@@ -27,8 +27,8 @@ from .runners import (
     DefaultEditor,
     DefaultRunners,
 )
-from .kitsu import KitsuProject
 from .site import Sites, Site, Synchronize, MinioFileUploader, MinioFileDownloader
+from .kitsu import KitsuAPIWrapper
 
 
 class LoginPageWidget(CustomPageWidget):
@@ -163,12 +163,130 @@ class GotoBookmarks(flow.Action):
         )
 
 
+class SiteChoiceValue(flow.values.ChoiceValue):
+
+    _parent = flow.Parent()
+
+    def choices(self):
+        return self.root().project().admin.sites.mapped_names()
+
+
+class RequestRevisionsAs(flow.Action):
+
+    path = flow.SessionParam("").ui(
+        placeholder="Revision oid pattern (e.g., /siren/films/siren/sequences/sq27/shots/*/departments/misc/files/*/history/revisions/v001)"
+    )
+    requesting_site_name = flow.Param("default", SiteChoiceValue).watched().ui(
+        label="Requesting site"
+    )
+    requested_site_name = flow.Param("default", SiteChoiceValue).watched().ui(
+        label="Site to query"
+    )
+
+    def allow_context(self, context):
+        current_site = self.root().project().get_current_site()
+        return context and current_site.request_files_from_anywhere.get()
+    
+    def child_value_changed(self, child_value):
+        if child_value in (self.requesting_site_name, self.requested_site_name):
+            msg = "<h2>Request revisions</h2>"
+            if self.requested_site_name.get() == self.requesting_site_name.get():
+                msg += (
+                    "<font color=#D5000D>"
+                    "Requested and requesting sites can't be the same."
+                    "</font>"
+                )
+
+            self.message.set(msg)
+
+    def get_buttons(self):
+        self.requested_site_name.set(self.requested_site_name.choices()[0])
+        return ["Request", "Close"]
+    
+    def run(self, button):
+        if button == "Close":
+            return
+        
+        requested_site_name = self.requested_site_name.get()
+        requesting_site_name = self.requesting_site_name.get()
+        
+        if requested_site_name == requesting_site_name:
+            return self.get_result(close=False)
+        
+        self.message.set("<h2>Request revisions</h2>")
+        
+        splitted_oids = self.path.get().split('*')
+        lvl = len(splitted_oids)
+        oids = []
+
+        def ls(oid):
+            if self.root().session().cmds.Flow.is_map(oid):
+                return self.root().session().cmds.Flow.get_mapped_oids(oid)
+
+            return list(map(
+                lambda rel: rel[0],
+                self.root().session().cmds.Flow.ls(oid)[0]
+            ))
+        
+        oids = [splitted_oids[0]]
+
+        for splitted_oid in splitted_oids[1:]:
+            updated_oids = []
+            if splitted_oid.startswith("/"):
+                splitted_oid = splitted_oid[1:]
+            
+            for oid in oids:
+                if oid.endswith("/"):
+                    oid = oid[:-1]
+                if not self.root().session().cmds.Flow.exists(oid):
+                    msg = self.message.get()
+                    msg += (
+                        "<font color=#D5000D>"
+                        f"Oid <b>{oid}</b> does not exist"
+                        "</font>"
+                    )
+                    self.message.set(msg)
+                    return self.get_result(close=False)
+
+                children = ls(oid)
+                
+                for child in children:
+                    updated_oids.append(child + "/" + splitted_oid)
+            
+            oids = updated_oids
+
+        processed_revisions = 0
+
+        for revision_oid in oids:
+            if not self.root().session().cmds.Flow.exists(revision_oid):
+                self.root().session().log_error(f"Revision {revision_oid} does not exist")
+                continue
+
+            request_action_oid = revision_oid + "/request_as"
+            self.root().session().cmds.Flow.set_value(
+                request_action_oid + "/requesting_site_name",
+                requesting_site_name
+            )
+            self.root().session().cmds.Flow.set_value(
+                request_action_oid + "/requested_site_name",
+                requested_site_name
+            )
+            processed_revisions += 1
+            self.root().session().cmds.Flow.run_action(request_action_oid, None)
+        
+        msg = "<h3>Request revisions</h3>"
+        msg += f"<b>{processed_revisions}</b> revision(s) processed."
+        self.message.set(msg)
+
+        return self.get_result(close=False)
+
+
 class Synchronization(flow.Object):
     ICON = ("icons.libreflow", "sync_arrow")
     
     _project = flow.Parent()
     synchronize_files = flow.Child(SynchronizeFiles)
-    bookmarks = flow.Child(GotoBookmarks)
+    request_revisions_as = flow.Child(RequestRevisionsAs)
 
     def summary(self):
         current_site = self._project.get_current_site()
@@ -207,6 +325,28 @@ class LogOut(flow.Action):
         )
 
 
+class KitsuConfig(flow.Object):
+
+    ICON = ('icons.libreflow', 'kitsu')
+
+    server_url = flow.Param().watched()
+    project_name = flow.Param().watched()
+    project_id = flow.Computed(
+        cached=True
+    )
+    gazu_api = flow.Child(KitsuAPIWrapper).ui(
+        hidden=True
+    )
+
+    def child_value_changed(self, child_value):
+        self.project_id.touch()
+
+    def compute_child_value(self, child_value):
+        if child_value is self.project_id:
+            data = self.gazu_api.get_project_data()
+            child_value.set(data["id"])
+
+
 class ProjectSettings(flow.Object):
     project_nice_name = flow.Param()
     project_thumbnail = flow.Param().ui(editor='textarea')
@@ -220,11 +360,12 @@ class Admin(flow.Object):
 
     current_site_name = flow.Computed(cached=True)
     root_dir = flow.Computed(cached=True)
-    kitsu_server_url = flow.Param("").watched()
+    # kitsu_server_url = flow.Param("").watched()
 
     project_settings = flow.Child(ProjectSettings)
     settings = flow.Child(ContextualView)
     users = flow.Child(Users)
+    kitsu = flow.Child(KitsuConfig)
     
     user_environment = flow.Child(UserEnvironment).ui(expanded=False)
     default_applications = flow.Child(DefaultRunners).ui(expanded=False).injectable()
@@ -284,7 +425,7 @@ class Admin(flow.Object):
         #     child_value.set(getpass.getuser())
 
 
-class Project(KitsuProject):
+class Project(flow.Object):
 
     log_out_action = flow.Child(LogOut).ui(label="Log out")
 
@@ -353,12 +494,13 @@ class Project(KitsuProject):
 
     def compute_child_value(self, child_value):
         if child_value is self._show_login_page:
-            valid_host = self.kitsu_api.host_is_valid()
+            kitsu_api = self.kitsu_api()
+            valid_host = kitsu_api.host_is_valid()
 
             if not valid_host:
                 logged_in = False
             else:
-                logged_in = self.kitsu_api.current_user_logged_in()
+                logged_in = kitsu_api.current_user_logged_in()
 
             if not logged_in:
                 try:
@@ -367,24 +509,27 @@ class Project(KitsuProject):
                     logged_in = False
                 else:
                     kitsu_config = json.load(f)
-                    self.kitsu_api.set_host(kitsu_config["kitsu_host"])
-                    self.kitsu_api.set_tokens(kitsu_config["tokens"])
+                    kitsu_api.set_host(kitsu_config["kitsu_host"])
+                    kitsu_api.set_tokens(kitsu_config["tokens"])
 
-                    logged_in = self.kitsu_api.current_user_logged_in()
+                    logged_in = kitsu_api.current_user_logged_in()
 
-            child_value.set(not self.kitsu_api.host_is_valid() or not logged_in)
+            child_value.set(not kitsu_api.host_is_valid() or not logged_in)
         else:
             super(Project, self).compute_child_value(child_value)
 
     def update_kitsu_host(self, server_url):
-        self.kitsu_api.set_server_url(server_url)
-        self.kitsu_api.set_host(server_url + "/api")
+        kitsu_api = self.kitsu_api()
+        kitsu_api.set_server_url(server_url)
+        kitsu_api.set_host(server_url + "/api")
 
-        return self.kitsu_api.host_is_valid()
+        return kitsu_api.host_is_valid()
 
     def log_in(self, kitsu_id, kitsu_password):
-        self.update_kitsu_host(self.admin.kitsu_server_url.get())
-        success = self.kitsu_api.log_in(kitsu_id, kitsu_password)
+        kitsu_api = self.kitsu_api()
+
+        self.update_kitsu_host(self.admin.kitsu.server_url.get())
+        success = kitsu_api.log_in(kitsu_id, kitsu_password)
 
         if success:            
             # Set current user
@@ -402,10 +547,10 @@ class Project(KitsuProject):
                 os.makedirs(user_settings_folder)
 
             # Save authentification tokens
-            tokens = self.kitsu_api.get_tokens()
+            tokens = kitsu_api.get_tokens()
             kitsu_config = {}
             kitsu_config["tokens"] = tokens
-            kitsu_config["kitsu_host"] = self.kitsu_api.get_host()
+            kitsu_config["kitsu_host"] = kitsu_api.get_host()
 
             if user_settings_folder is not None:
                 with open("%s/kitsu_config.json" % user_settings_folder, "w+") as f:
@@ -414,7 +559,16 @@ class Project(KitsuProject):
         return success
     
     def log_out(self):
-        self.kitsu_api.log_out()
+        self.kitsu_api.log_out()()
+    
+    def kitsu_id(self):
+        return self.admin.kitsu.project_id.get()
+
+    def kitsu_name(self):
+        return self.admin.kitsu.project_name.get()
+    
+    def kitsu_api(self):
+        return self.admin.kitsu.gazu_api
 
     def get_contextual_view(self, context_name):
         if context_name == "settings":
@@ -504,12 +658,16 @@ class Project(KitsuProject):
     def get_factory(self):
         self.ensure_runners_loaded()
         return self._RUNNERS_FACTORY
+    
+    def touch(self):
+        super(Project, self).touch()
+        self.update_user_environment()
 
     def _fill_ui(self, ui):
         if self._RUNNERS_FACTORY is None:
             self.ensure_runners_loaded()
 
-        self.update_user_environment()
+        self.touch()
 
         if self._show_login_page.get():
             ui["custom_page"] = "libreflow.baseflow.LoginPageWidget"
